/*************************************************************************
ID: popa3d/virtual.c

Portions Copyright 2002, Bulent Yilmaz

Part of Microbrew Microbrew Message Center
Multi-server Mailer Suite

http://www.microbrew.org/products/mmc/

Microbrew Message Center is free software; you can redistribute it and/or  
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

Microbrew Message Center is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License
along with MicroSieve; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

A full copy of the GNU General Public License can be found at:
http://www.gnu.org/licenses/gpl.txt
**************************************************************************/

/*
 * Virtual domain support.
 */

#include "params.h"

#if POP_VIRTUAL

#define _XOPEN_SOURCE 4
#define _XOPEN_SOURCE_EXTENDED
#define _XOPEN_VERSION 4
#define _XPG4_2
#define __USE_SVID
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "../src/popb4relay.h"

#ifndef NAME_MAX
#define NAME_MAX			255
#endif

#define PREFIX "/usr/local/mmc"
#define IPDOMMAP PREFIX "/etc/ipdommap.conf"

extern int log_error(char *s);
extern struct passwd *fgetpwent (FILE *);

char *virtual_domain;
char *virtual_spool;

int virtual_startup(void)
{
	return 0;
}

unsigned long client_lookup() {

    int length;
    struct sockaddr_in sin;
    
    length = sizeof(sin);
    if (getsockname(0, (struct sockaddr *)&sin, &length)) {
        if (errno == ENOTSOCK) return 0;
        log_error("getsockname");
        return 0;
    }
    
    if (length != sizeof(sin) || sin.sin_family != AF_INET) return 0;
    
    return sin.sin_addr.s_addr;

}

static char *lookup(void)
{
        struct sockaddr_in sin;
        int length;
        char *ip, *domain;
        static char **ipaddr=NULL, **domainmap=NULL;
        FILE *ipdommap=NULL;
        char readbuf[2048];
        static int maps;
        int i;

        if(ipaddr==NULL) {
            maps=0;
            if((ipdommap = fopen(IPDOMMAP, "r")) != NULL) {
                do {
                    fgets(readbuf, 2047, ipdommap);
                    if(readbuf[0] != '#') {
                        ip = strtok(readbuf, "\t ");
                        domain = strtok(NULL, "\r\n\t ");
                        if((ip != NULL) && (domain != NULL)) {
                            ipaddr=realloc(ipaddr, ((maps+1)*sizeof(char *)) );
                            *(ipaddr+maps)=strdup(ip);
                            domainmap=realloc(domainmap, ((maps+1)*sizeof(char *)) );
                            *(domainmap+maps)=strdup(domain);
                            maps++;
                        }
                    }
                } while(!feof(ipdommap));
                fclose(ipdommap);
            }
        }

        length = sizeof(sin);
        if (getsockname(0, (struct sockaddr *)&sin, &length)) {
                if (errno == ENOTSOCK) return "";
                log_error("getsockname");
                return NULL;
        }
        if (length != sizeof(sin) || sin.sin_family != AF_INET) return NULL;

        ip = inet_ntoa(sin.sin_addr);
        
        for(i=0; i<maps; i++) {
            if(!strcmp( ip, *(ipaddr+i) )) return *(domainmap+i);
        }
        
        log_error("IP to Domain map");
        
        return NULL;
}

static int is_valid_user(char *user)
{
	unsigned char *p;

/* This is pretty liberal, but we're going to use direct syscalls only,
 * and they have to accept all the printable characters */
	for (p = (unsigned char *)user; *p; p++)
		if (*p < ' ' || *p > 0x7E || *p == '/') return 0;

	if (p - (unsigned char *)user > NAME_MAX) return 0;

	return 1;
}

struct passwd *virtual_userpass(char *user, char *pass, int *known)
{
	struct passwd *pw;
	struct stat stat;
	char *address, *pathname, *tempuser;
	FILE *passfile;
	int match=0;
	char *myuser;
	unsigned long cli_ip;
	
	/* Start out with a good paranoid set of defaults. */
	
	*known = 0;
        virtual_domain = NULL;
        
//	log_error("virtual_userpass");

        /* Check to make sure that ther are no malicious characters in the 
         * supplied username. */

	if (!is_valid_user(user)) {
		log_error("Invalid user");
		user = "INVALID";
		return NULL;
	}

	/* Copy username data since strtok() will destroy my data as a free
	 * gift.  Pull out the username@domain format.  */

	if((tempuser = strdup(user)) != NULL) {
		myuser = strtok(tempuser, "@\r\n\0");
		address = strtok(NULL, "@\r\n\0");
	} else {
		log_error("Input username strdup");
		user = "INVALID";
		return NULL;
	}
//      user = "INVALID";

        /* Make sure that both the username and the domain are valid. */

        if(myuser == NULL) {
            /* No username, hell no. */
        
            log_error("Invalid username format");
            free(tempuser);
            return NULL;
        } else {
            /* We have a username, do we have a domain? */
            
            if(address == NULL) {
                /* No domain.  Maybe there's an IP to Domain map. */
                
                address = lookup();
                
                /* If there is no map, we don't know who's mail it is. */
                if(address == NULL) {
                    log_error("Invalid domain format");
                    free(tempuser);
                    return NULL;
                }
            }
        }
#if 0
	if ((!address)||(!myuser)) {
		log_error("Invalid username format");
		free(tempuser);
		return NULL;
	}
#endif
	
        /* Make copy of the data I want and free the memory.  Be pedantic 
         * about it. */ 

	if( (myuser=strdup(myuser)) == NULL ) {
		log_error("strdup(myuser)");
		free(tempuser);
		return NULL;
	}	
	if( (address=strdup(address)) == NULL ) {
		log_error("strdup(address)");
		free(myuser);
		free(tempuser);
		return NULL;
	}
	free(tempuser);
	

	/* This "can't happen", but is just too critical to not check 
	 * explicitly. */

	
	if (strchr(address, '/') || strchr(myuser, '/')) {
		free(myuser);
		free(address);
		return NULL;
	}

        /* Allocate memory and determine my path to the virtual domain. */

	pathname = malloc(strlen(VIRTUAL_HOME_PATH) + strlen(address) + strlen(VIRTUAL_AUTH_PATH) + strlen(myuser) + 4);
	if (pathname == NULL) {
		log_error("Pathname malloc()");
		free(myuser);
		free(address);
		return NULL;
	}
	sprintf(pathname, "%s/%s", VIRTUAL_HOME_PATH, address);

        /* Check to make sure that my path to the virtual domain actually 
         * exists. */

	if (lstat(pathname, &stat)) {
		log_error("lstat");
		free(myuser);
		free(address);
		free(pathname);
		return NULL;
	}

        /* Determine my path to the auth file for this user. */

	sprintf(pathname, "%s/%s/%s/%s", VIRTUAL_HOME_PATH, address, VIRTUAL_AUTH_PATH, myuser);

	/* Open up the file and see if we can get a '/etc/passwd' style entry. 
	 * for the specified user. */

	if ( (passfile = fopen(pathname, "r")) ) {
	
		match = 1;
		do {
			pw = fgetpwent(passfile);
			if( pw == NULL) break;
			
		} while ( (( match = strcmp(pw->pw_name, myuser) )!=0) && (!feof(passfile)) );

		fclose(passfile);
		free(pathname);
		free(myuser);
		
		if(match) {
			log_error("Missing auth data");
			free(address);
			return NULL;
		}
			
	} else {	
		log_error("Auth file open");
		free(address);
		free(myuser);
		free(pathname);
		return NULL;
	}
	
	/* It exists.  Proceed to set my virtual_domain to that address.  
         * At this point, address shouldn't be freed, and hopefully it will
         * get done later. */

	virtual_domain = address;

        /* Compare the supplied password with the password from the file. */

	if(!strcmp( crypt(pass, pw->pw_passwd), pw->pw_passwd )) {

		/* The user authenticates, lets point virtual_spool at the 
		 * user's mailbox.  Again, be pedantic. */

		if( (virtual_spool = malloc(strlen(VIRTUAL_HOME_PATH) + strlen(virtual_domain) + strlen(VIRTUAL_SPOOL_PATH) + 3)) == NULL) {
			log_error("virtual_spool malloc");
			free(address);
			virtual_domain = NULL;
			return NULL;
		}	
		sprintf(virtual_spool, "%s/%s/%s", VIRTUAL_HOME_PATH, virtual_domain, VIRTUAL_SPOOL_PATH);

		/* Wipe passwords from memory. */
		memset(pw->pw_passwd, 0, strlen(pw->pw_passwd));

                /* Add client IP to the POP before relay server. */
                cli_ip=client_lookup();
                if(cli_ip != 0) popb4relay_add(cli_ip);
		
		/* Set the mailbox name (user) correctly. */
		strncpy(user, pw->pw_name, strlen(pw->pw_name));
		*(user+strlen(pw->pw_name))=0;
		
		/* Return the data. */
		*known = 1;		
		return pw;		
	} else {
		log_error("password mismatch");
		free(address);
		free(myuser);
		virtual_domain = NULL;
		return NULL;
	}
	
	/* Just in case: */
	
	return NULL;
	
}

#endif
